@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
+#include <pthread.h>
 
 #include <xine/xine_internal.h>
 #include <xine/xineutils.h>
@@ -85,6 +86,7 @@ typedef struct demux_xvdr_s {
   input_plugin_t       *input;
 
   ts_data_t            *ts_data;        /* MPEG-TS stuff */
+  pthread_mutex_t       ts_data_seek_lock;
 
   int64_t               last_pts[2];
   int64_t               last_vpts;
@@ -93,6 +95,10 @@ typedef struct demux_xvdr_s {
   uint32_t              audio_type;
   uint32_t              subtitle_type;
 
+  /* recovery point detection and initial buffering */
+  unsigned              wait_recovery_point;
+  unsigned              I_frames, P_frames, B_frames;
+
   /* current buf_element */
   int64_t               pts;
   int64_t               dts;
@@ -228,7 +234,6 @@ static void check_newpts(demux_xvdr_t *this, buf_element_t *buf, int video )
     if (this->send_newpts || (this->last_pts[video] && abs(diff)>WRAP_THRESHOLD)) {
 
       LOGVERBOSE("New PTS: %"PRId64" (%s)", buf->pts, video ? "VIDEO" : "AUDIO");
-
       if (this->buf_flag_seek) {
         _x_demux_control_newpts(this->stream, buf->pts, BUF_FLAG_SEEK);
         this->buf_flag_seek = 0;
@@ -324,6 +329,7 @@ static void track_audio_stream_change(demux_xvdr_t *this, buf_element_t *buf)
         b->type = BUF_CONTROL_START;
         b->decoder_flags = BUF_FLAG_GAPLESS_SW;
         this->audio_fifo->put(this->audio_fifo, b);
+        LOGVERBOSE("posted BUF_CONTROL_START to audio fifo");
       }
     }
 
@@ -357,6 +363,7 @@ static void demux_xvdr_fwd_buf(demux_xvdr_t *this, buf_element_t *buf)
   /* decoder flush --> video_fifo */
   if (buf->type == BUF_CONTROL_FLUSH_DECODER) {
     /* append sequence end code */
+    /* note: ts is already flushed (for all ctrl bufs) */
     post_sequence_end (this->video_fifo, this->video_type);
     this->video_fifo->put (this->video_fifo, buf);
     return;
@@ -382,6 +389,91 @@ static void demux_xvdr_fwd_buf(demux_xvdr_t *this, buf_element_t *buf)
   buf->free_buffer (buf);
 }
 
+/*
+ * Frame counting for initial buffering
+ */
+
+static void count_frames(demux_xvdr_t *this, buf_element_t *buf)
+{
+  uint8_t type = NO_PICTURE;
+
+  if (!buf) {
+    this->I_frames = this->P_frames = this->B_frames = 0;
+    return;
+  }
+  if (this->I_frames > 4 || buf->size < 4) {
+    return;
+  }
+
+  if (buf->type == BUF_VIDEO_MPEG) {
+    type = mpeg2_get_picture_type(buf->content, buf->size);
+  } else if (buf->type == BUF_VIDEO_H264) {
+    type = h264_get_picture_type(buf->content, buf->size);
+  }
+
+  if (type == NO_PICTURE)
+    return;
+
+  switch (type) {
+    case I_FRAME: this->I_frames++; break;
+    case P_FRAME: if (this->I_frames > 0) this->P_frames++; break;
+    case B_FRAME: if (this->I_frames > 0) this->B_frames++; break;
+    default: break;
+  }
+
+  LOGVERBOSE("frames: I %d P %d B %d\n", this->I_frames, this->P_frames, this->B_frames);
+  int data[] = {this->I_frames, this->P_frames, this->B_frames, 0};
+  this->stream->input_plugin->get_optional_data(this->stream->input_plugin, data,
+                                                INPUT_OPTIONAL_DATA_XVDR_FRAMES);
+}
+
+static int _detect_recovery_point(buf_element_t *buf)
+{
+  uint8_t *p = buf->content;
+
+  if (buf->size <= 4)
+    return 0;
+
+  if (buf->type == BUF_VIDEO_MPEG) {
+
+    if (mpeg2_is_sequence_header(buf->content, buf->size)) {
+      LOGDBG("MPEG2 sequence header detected");
+      return 1;
+    }
+
+    if (mpeg2_get_picture_type(buf->content, buf->size) == I_FRAME) {
+      LOGMSG("MPEG2 I-frame without sequence header ?");
+    }
+
+    LOGVERBOSE("skipping MPEG2 block: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
+               p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8], p[9], p[10]);
+    return 0;
+  }
+
+  if (buf->type == BUF_VIDEO_H264) {
+    /* scan video packet for sequence parameter set */
+    for (int i = 0; i < buf->size - 4; i++) {
+      if (p[i] == 0 && p[i + 1] == 0 && p[i + 2] == 1) {
+        if ((p[i + 3] & 0x1f) == NAL_SPS) {
+          LOGDBG("detect_recovery_point: found SPS at %d\n", i);
+          return 1;
+        } else if ((p[i + 3] & 0x1f) == NAL_PPS) {
+          LOGDBG("detect_recovery_point: found PPS at %d\n", i);
+        }
+      }
+    }
+    if (h264_get_picture_type(buf->content, buf->size) == I_FRAME) {
+      LOGMSG("H.264 I-frame without SPS ?");
+    }
+    LOGVERBOSE("skipping H.264 block: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x",
+               p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8], p[9], p[10]);
+    return 0;
+  }
+
+  /* TODO: not implemented for H.265 */
+  return 1;
+}
+
 /*
  * demux_xvdr_parse_ts()
  *
@@ -405,11 +497,12 @@ static unsigned int vtype_to_xine_buf_type(unsigned vtype)
 
 static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
 {
+  pthread_mutex_lock(&this->ts_data_seek_lock);
+
   if (!this->ts_data)
     this->ts_data = calloc(1, sizeof(ts_data_t));
 
   ts_data_t     *ts_data  = this->ts_data;
-  fifo_buffer_t *src_fifo = buf->source;
 
   while (buf->size >= TS_SIZE) {
 
@@ -444,12 +537,14 @@ static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
         /* PMT changed, reset ts->es converters */
         LOGDBG("PMT changed, resetting demuxer");
         ts_data_ts2es_init(&ts_data, this->stream->video_fifo, this->stream->audio_fifo);
+        count_frames(this, NULL);
 
         this->video_type = vtype_to_xine_buf_type(ts_data->pmt.video_type);
-        if (!this->video_type) {
+        if (ts_data->pmt.video_type && !this->video_type) {
           LOGMSG("unsupported video codec 0x%02x detected (no support in xine-lib ?)", ts_data->pmt.video_type);
           ts_data->pmt.video_pid = INVALID_PID;
         }
+        this->wait_recovery_point = (this->video_type != 0);
 
         /* Inform UI of channels changes */
         xine_event_t event;
@@ -462,11 +557,26 @@ static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
     /* demux video */
     else if (ts_pid == ts_data->pmt.video_pid) {
       if (ts_data->video) {
-        buf_element_t *vbuf     = ts2es_put(ts_data->video, buf->content, src_fifo);
+        buf_element_t *vbuf     = ts2es_put(ts_data->video, buf->content);
         while (vbuf) {
           buf_element_t *next = vbuf->next;
+          vbuf->next = NULL;
+          if (this->wait_recovery_point) {
+            this->wait_recovery_point = (this->wait_recovery_point + 1) * (!_detect_recovery_point(vbuf));
+            if (this->wait_recovery_point > 50) {
+              LOGMSG("detect_recovery_point: WARNING: Skipped 50+ frames !\n");
+              this->wait_recovery_point = 0;
+            }
+            if (this->wait_recovery_point) {
+              if (!next && !(vbuf->decoder_flags & BUF_FLAG_FRAME_END))
+                ts2es_reset(ts_data->video);
+              vbuf->free_buffer(vbuf);
+              vbuf = next;
+              continue;
+            }
+          }
           check_newpts( this, vbuf, PTS_VIDEO );
-
+          count_frames(this, vbuf);
           this->stream->video_fifo->put(this->stream->video_fifo, vbuf);
           vbuf = next;
         }
@@ -475,15 +585,17 @@ static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
 
     /* demux audio */
     else {
+      if (!this->wait_recovery_point) {
       int i/*, done = 0*/;
       for (i=0; i < ts_data->pmt.audio_tracks_count; i++)
         if (ts_pid == ts_data->pmt.audio_tracks[i].pid) {
           if (ts_data->audio[i]) {
-            buf_element_t *abuf = ts2es_put(ts_data->audio[i], buf->content, src_fifo);
+            buf_element_t *abuf = ts2es_put(ts_data->audio[i], buf->content);
             while (abuf) {
               buf_element_t *next = abuf->next;
               check_newpts( this, abuf, PTS_AUDIO );
               track_audio_stream_change (this, abuf);
+              abuf->next = NULL;
 
               this->stream->audio_fifo->put(this->stream->audio_fifo, abuf);
               abuf = next;
@@ -492,6 +604,7 @@ static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
           /*done = 1;*/
           break;
         }
+      }
 #if 0
       /* demux subtitles */
       if (!done)
@@ -517,6 +630,8 @@ static void demux_xvdr_parse_ts (demux_xvdr_t *this, buf_element_t *buf)
     buf->size    -= TS_SIZE;
   }
 
+  pthread_mutex_unlock(&this->ts_data_seek_lock);
+
   buf->free_buffer(buf);
 }
 
@@ -1009,6 +1124,7 @@ static int32_t parse_video_stream(demux_xvdr_t *this, uint8_t *p, buf_element_t
   buf->size      = this->packet_len;
 
   check_newpts( this, buf, PTS_VIDEO );
+  count_frames( this, buf );
 
   this->video_fifo->put (this->video_fifo, buf);
 
@@ -1056,8 +1172,10 @@ static void demux_xvdr_parse_pes (demux_xvdr_t *this, buf_element_t *buf)
   this->stream_id  = p[3];
 
   if (this->ts_data) {
+    pthread_mutex_lock(&this->ts_data_seek_lock);
     ts_data_flush(this->ts_data);
     ts_data_dispose(&this->ts_data);
+    pthread_mutex_unlock(&this->ts_data_seek_lock);
   }
 
   if (IS_VIDEO_PACKET(p)) {
@@ -1094,7 +1212,9 @@ static void demux_xvdr_parse_pack (demux_xvdr_t *this)
   if (!buf) {
     if (errno == EINTR) {
       LOGVERBOSE("input->read_block() was interrupted");
+      pthread_mutex_lock(&this->ts_data_seek_lock);
       ts_data_flush(this->ts_data);
+      pthread_mutex_unlock(&this->ts_data_seek_lock);
     } else if (errno != EAGAIN) {
       LOGDBG("DEMUX_FINISHED (input returns NULL with error)");
       this->status = DEMUX_FINISHED;
@@ -1106,7 +1226,14 @@ static void demux_xvdr_parse_pack (demux_xvdr_t *this)
   /* If this is not a block for the demuxer, pass it
    * straight through. */
   if (buf->type != BUF_DEMUX_BLOCK) {
-    ts_data_flush (this->ts_data);
+    pthread_mutex_lock(&this->ts_data_seek_lock);
+    if (buf->type == BUF_CONTROL_RESET_DECODER) {
+      count_frames(this, NULL);
+      ts_data_seek (this->ts_data);
+    } else {
+      ts_data_flush (this->ts_data);
+    }
+    pthread_mutex_unlock(&this->ts_data_seek_lock);
     demux_xvdr_fwd_buf (this, buf);
     return;
   }
@@ -1114,6 +1241,7 @@ static void demux_xvdr_parse_pack (demux_xvdr_t *this)
   p = buf->content; /* len = this->blocksize; */
   buf->decoder_flags = 0;
 
+  if (LIKELY(buf->size > 5)) {
   if (DATA_IS_TS(p)) {
     demux_xvdr_parse_ts (this, buf);
     return;
@@ -1122,6 +1250,7 @@ static void demux_xvdr_parse_pack (demux_xvdr_t *this)
     demux_xvdr_parse_pes (this, buf);
     return;
   }
+  }
 
   LOGMSG("Header %02x %02x %02x (should be 0x000001 or 0x47)", p[0], p[1], p[2]);
   buf->free_buffer (buf);
@@ -1149,6 +1278,8 @@ static void demux_xvdr_dispose (demux_plugin_t *this_gen)
 
   ts_data_dispose(&this->ts_data);
 
+  pthread_mutex_destroy(&this->ts_data_seek_lock);
+
   free (this);
 }
 
@@ -1157,10 +1288,12 @@ static int demux_xvdr_get_status (demux_plugin_t *this_gen)
   demux_xvdr_t *this = (demux_xvdr_t *) this_gen;
 
   if (this->status != DEMUX_OK) {
+    pthread_mutex_lock(&this->ts_data_seek_lock);
     if (this->ts_data) {
       LOGMSG("demux_xvdr_get_status(): status != DEMUX_OK. -> freeing ts_data");
       ts_data_dispose(&this->ts_data);
     }
+    pthread_mutex_unlock(&this->ts_data_seek_lock);
   }
 
   return this->status;
@@ -1200,7 +1333,11 @@ static int demux_xvdr_seek (demux_plugin_t *this_gen,
   this->audio_type    = 0;
   this->subtitle_type = 0;
   this->bih_posted    = 0;
+
+  pthread_mutex_lock(&this->ts_data_seek_lock);
+  count_frames(this, NULL);
   ts_data_dispose(&this->ts_data);
+  pthread_mutex_unlock(&this->ts_data_seek_lock);
 
   if (!playing) {
 
@@ -1211,7 +1348,20 @@ static int demux_xvdr_seek (demux_plugin_t *this_gen,
   } else {
     this->buf_flag_seek = 1;
     this->last_vpts     = INT64_C(-1);
+
+    LOGVERBOSE("demux_xvdr_seek: flushing engine...");
+
+    /* Make sure this sends DISC_STREAMSTART. */
+    int gs = xine_get_param (this->stream, XINE_PARAM_GAPLESS_SWITCH);
+    if (gs)
+      xine_set_param (this->stream, XINE_PARAM_GAPLESS_SWITCH, 0);
+    _x_demux_control_start (this->stream);
+    if (gs)
+      xine_set_param (this->stream, XINE_PARAM_GAPLESS_SWITCH, gs);
+
     _x_demux_flush_engine(this->stream);
+
+    LOGVERBOSE("demux_xvdr_seek: engine flushed");
   }
 
   return this->status;
@@ -1274,6 +1424,8 @@ static demux_plugin_t *demux_xvdr_open_plugin (demux_class_t *class_gen,
   detect_video_decoders(this);
   detect_audio_decoders(this);
 
+  pthread_mutex_init(&this->ts_data_seek_lock, NULL);
+
   return &this->demux_plugin;
 }
 
